# -*- coding: utf-8 -*-

# $Id:$

"""Python djira's client.

This module serves both as an example for client implementers and as a
test for the API usability from the client perspective.

"""

from __future__ import print_function
from __future__ import unicode_literals

import json
from urllib.error import HTTPError
from urllib.error import URLError
from urllib.parse import urlencode
from urllib.request import urlopen


class _Proxy(object):

    # NOTE: I use slots as a way to discourege tinkering with the
    # instance state. Once instantiated a proxy must be considered
    # read-only so that it can be reused.
    __slots__ = ("_url", "_endpoint", "_full_url", "_url_opener")

    def __init__(self, url, endpoint, url_opener=urlopen):
        if not url.endswith("/"):
            url = url + "/"
        if not endpoint.endswith("/"):
            endpoint = endpoint + "/"
        self._url = url
        self._endpoint = endpoint
        self._full_url = url + endpoint
        self._url_opener = url_opener

    def __call__(self, **kwargs):
        url = self._make_url(self._full_url, kwargs)
        try:
            return self._read_from_url(url, self._url_opener)
        except HTTPError as e:
            # HACK: log instead of printing
            print("HTTPError, status {}".format(e.code))
            return None  # HACK: must raise some exception
        except URLError as e:
            print("URLError, reason {}".format(e.reason))
            return None

    @staticmethod
    def _make_url(url, param_dict):
        if param_dict:
            # TODO: test with non ASCII input
            qs = urlencode(list(param_dict.items()), doseq=True)
            url = url + "?" + qs  # HACK: that don't look good
        return url

    @staticmethod
    def _read_from_url(url, opener):
        with opener(url) as response:
            if response.status != 200:
                raise RuntimeError("status != 200: {}".format(response.status))
            if response.getheader("Content-Type") != "application/json":
                raise RuntimeError(
                    "unsupported content-type: '{}'".format(
                        response.getheader("Content-Type")
                    )
                )
            return json.loads(response.read())


class DjiraClient(object):

    MIN_API_VERSION_SUPPORTED = (1, 0)

    def __init__(self, url):
        self.url = url
        self._proxies = {}

    def __getattr__(self, name):
        if name not in self._proxies:
            self._proxies[name] = _Proxy(self.url, name)
        return self._proxies[name]

    def check_minimum_api_version(self):
        version = tuple(self.__version__())
        return self.MIN_API_VERSION_SUPPORTED <= version



# -*- coding: utf-8 -*-

# $Id:$

"""This modules defines the plugin API.

"""

import pluggy

from .common import PROJECT_NAME

hookspec = pluggy.HookspecMarker(PROJECT_NAME)


@hookspec
def initialize():
    """Initialize the plugin.

    Function called when djira app is initialized (method ``ready``).
    A plugin can implement this hook if it requires some
    initialization.

    """


@hookspec
def get_endpoints():
    """Returns the endpoints defined by the plugin.

    Returns a list of ``EndPoint`` instances.

    """



# -*- coding: utf-8 -*-

# $Id:$

from __future__ import print_function
from __future__ import unicode_literals

from contextlib import contextmanager
import unittest

from .. import validators as V


class _BaseTestCase(unittest.TestCase):

    @contextmanager
    def assertValidationFails(self, regex=None):
        if regex is None:
            manager = self.assertRaises(V.SchemaError)
        else:
            manager = self.assertRaisesRegexp(V.SchemaError, regex)
        with manager:
            yield


class TextMayBeNone(_BaseTestCase):

    def setUp(self):
        def is_even(v):
            if v % 2:
                raise V.SchemaError("not even")
            return v

        self.validator = is_even
        self.new_validator = V.maybe_None(is_even)

    def test_decorated_validator_accepts_None(self):
        # make sure validator(None) errors, just in case
        with self.assertRaises(TypeError):
            self.validator(None)

        self.assertIsNone(self.new_validator(None))

    def test_decorated_validator_behaves_the_same_for_not_None(self):
        # just to be safe
        self.assertEqual(self.validator(2), 2)
        with self.assertRaises(V.SchemaError):
            self.validator(3)

        self.assertEqual(self.new_validator(2), 2)
        with self.assertRaises(V.SchemaError):
            self.new_validator(3)


class TestBounded(_BaseTestCase):

    # factory

    def test_not_callable_getter_raises_TypeError(self):
        with self.assertRaisesRegexp(TypeError, r"'getter' must be callable"):
            V.bounded(min=1, getter=42)

    def test_missing_min_and_max_raises_ValueError(self):
        with self.assertRaisesRegexp(ValueError, r"must specify 'min' or 'max'"):
            V.bounded()

    # validator

    def test_min_value(self):
        validator = V.bounded(min=10)
        print(validator.__doc__)
        validator(10)
        validator(11)
        with self.assertValidationFails(r"value too small"):
            validator(9)

    def test_max_value(self):
        validator = V.bounded(max=10)
        validator(10)
        validator(9)
        with self.assertValidationFails(r"value too big"):
            validator(11)

    def test_both_bounds(self):
        validator = V.bounded(min=5, max=10)
        validator(5)
        validator(7)
        validator(10)
        with self.assertValidationFails(r"value too small"):
            validator(4)
        with self.assertValidationFails(r"value too big"):
            validator(12)

    def test_getter(self):
        validator = V.bounded(max=5, getter=len)
        validator("asdf")
        validator("asdfg")
        with self.assertValidationFails(r"value too big"):
            validator("asdfgh")



# -*- coding: utf-8 -*-

# $Id:$

# sample plugin for testing/demonstration purpose

from __future__ import print_function
from __future__ import unicode_literals

from ..common import EndPoint
from ..common import hookimpl
from .. import schema as S


@hookimpl
def initialize():
    print("demo: initialize")


@hookimpl
def get_endpoints():
    return (
        EndPoint(get_models_names),
        EndPoint(get_model_info, request_schema=get_model_info_schema),
    )


def get_models_names():
    """Return a list with the models names."""
    return ["foo", "bar", "baz"]


get_model_info_schema = S.Schema(
    schema=dict(
        model_id=S.String(doc="Model name."),
    ),
)


def get_model_info(model_id):
    """Return a dict with info about the given model."""
    return {
        "name": model_id,
        "verbose_name": "Some descriptive text",
        "fields": [
            {"name": "my_field", "type": "int"},
        ]
    }



# -*- coding: utf-8 -*-

# $Id:$

from __future__ import print_function
from __future__ import unicode_literals

import importlib
import json
import os
import sys

import django
from django.conf import settings

from ..common import EndPoint
from ..common import hookimpl


_cached_settings = None


@hookimpl
def get_endpoints():
    return (
        EndPoint(get_system_info),
    )


def get_system_info():
    """Return a dictionary with assorted system information.

    """
    settings_module_name = os.environ.get("DJANGO_SETTINGS_MODULE", None)
    if settings_module_name:
        try:
            settings_module = importlib.import_module(settings_module_name)
        except ModuleNotFoundError:
            project_root = None
        else:
            settings_module_path = settings_module.__file__
            if settings_module_path.endswith(".pyc"):
                settings_module_path = settings_module_path[:-1]
            project_root = os.path.dirname(os.path.dirname(settings_module.__file__))
    else:
        project_root = None

    return {
        "django": django.__path__[0],
        "django_project_root": project_root,
        "django_settings": _get_settings(),
        "django_settings_module": settings_module_name,
        "django_settings_path": settings_module_path,
        "django_version": django.VERSION,
        "python": sys.executable,
        "python_version": list(sys.version_info),
    }


def _get_settings():
    """Returns a dictionary with the settings.

    The returned dictionary i guaranteed to be json serializable, the
    values that aren't are converted to string with ``repr``.

    The dictionary is initialized in the first call and reused in
    subsequent calls.

    """
    global _cached_settings

    if _cached_settings is None:
        res = {}
        for i in dir(settings):
            if i.startswith("_"):
                continue
            v = getattr(settings, i)
            try:
                json.dumps(v)
            except TypeError:
                v = repr(v)
            res[i] = v
        _cached_settings = res
    return _cached_settings



# -*- coding: utf-8 -*-

# $Id: api.py 1068 2015-04-02 14:43:59Z alex $

from .common import EndPoint
from .common import hookimpl
from . import schema


__all__ = [
    "EndPoint",
    "hookimpl",
    "schema",
]



# -*- coding: utf-8 -*-

# $Id:$

# URL test patterns for djira. Use this file to ensure a consistent
# set of URL patterns are used when running unit tests. This test_urls
# module should be referred to by your test class.

from django.conf.urls import include
from django.conf.urls import url


urlpatterns = [
    # url(r"^accounts/", include("django.contrib.auth.urls")),
    # url(r"^formsng/", include("hera_formsng.urls")),
    # url(r"^utils/", include("hera_utils.urls")),
    url(r"^app/", include("djira.urls")),
]



# -*- coding: utf-8 -*-

# $Id:$

from functools import partial

from django.conf.urls import url

from djira import views


urlpatterns = [
    url(r"^$", partial(views.dispatcher, name="__list__"), name="djira_root_view"),
    url(r"^(?P<name>\w+)/$", views.dispatcher, name="djira_dispatcher_view")
]



# -*- coding: utf-8 -*-

# $Id:$

"""Validators and validator factories.

A validator is a callable that receives a value, perfoms some check on
it and either returns the value (maybe modified) or raises an
``SchemaError`` exception.

"""

from __future__ import print_function
from __future__ import unicode_literals

import functools


class SchemaError(Exception):
    """Validation error.

    When raising, pass an string describing the error. Don't
    capitalize the sentence and omit the final dot.

       >>> raise SchemaError("not a valid integer")

    """
    pass


def maybe_None(f):
    """Decorates a validator as accepting the value ``None``.

    When writing a validator usually it's advised to check for
    ``None`` values, otherwise bad things will happen:

    .. code-block:: pycon

       >>> from djira.validators import SchemaError
       >>> def is_odd(v):
       ...     if v % 2 == 0:
       ...         raise SchemaError("not odd")
       ...     return v
       ...
       >>> is_odd(3)
       3
       >>> is_odd(4)
       Traceback (most recent call last):
         ...
       djira.validators.SchemaError: not odd
       >>> is_odd(None)
       Traceback (most recent call last):
         ...
       TypeError: unsupported operand type(s) for %: 'NoneType' and 'int'

    This decorator returns a new validator that will handle ``None``
    values safely:

    .. code-block:: pycon

       >>> from djira.validators import maybe_None
       >>> new_is_odd = maybe_None(is_odd)
       >>> new_is_odd(3)
       3
       >>> new_is_odd(4)
       Traceback (most recent call last):
         ...
       djira.validators.SchemaError: not odd
       >>> new_is_odd(None)
       None

    """
    @functools.wraps(f)
    def validator(v):
        if v is None:
            return None
        return f(v)
    return validator


def bounded(min=None, max=None, getter=lambda x: x):
    """Check boundaries.

    This factory returns a validator that checks the boundaries for
    some property of ``value``, the value itself by default:

    .. code-block:: pycon

       >>> from djira.validators import bounded
       >>> validator = bounded(max=10)
       >>> validator(5)
       5
       >>> validator(11)
       Traceback (most recent call last):
         ...
       SchemaError: value too big


    It can check other properties, just specify a ``getter``:

    .. code-block:: pycon

       >>> len_validator = bounded(min=1, max=10, getter=len)
       >>> len_validator("hello")
       'hello'
       >>> len_validator("hello world")
       Traceback (most recent call last):
         ...
       SchemaError: value too big


    The validator can be applied to any type implementing comparison
    methods, not just numbers::

    .. code-block:: pycon

       >>> validator = bounded(min="ab", max="yz")
       >>> validator("abc")
       'abc'
       >>> validator("zz")
       Traceback (most recent call last):
         ...
       djira.validators.SchemaError: value too big


    :param min: minimum value. No minimum by default.

    :param max: maximum value. No maximum by default.

    :param callable getter: callable thar returns the value of the
      property. By default returns the value itself.

    :returns: a validator

    :raises TypeError: if ``getter`` is not callable.

    :raises ValueError: if neither ``min`` nor ``max``

    """

    if not callable(getter):
        raise TypeError("'getter' must be callable")
    if min is None and max is None:
        raise ValueError("must specify 'min' or 'max'")

    def validator(v):
        value = getter(v)
        if min is not None and value < min:
            raise SchemaError("value too small")
        if max is not None and value > max:
            raise SchemaError("value too big")
        return v

    return validator



# -*- coding: utf-8 -*-

# $Id: $

# Settings to be used when running unit tests
# python manage.py test --settings=djira.test_settings djira

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "",
        "USER": "",
        "PASSWORD": "",
        "HOST": "",
        "PORT": "",
    }
}

SECRET_KEY = "Tests don't require a really s3cr3t k3y ;)"

INSTALLED_APPS = (
    # Put any other apps that your app depends on here

    #BEGIN:installed_apps
    # "django.contrib.auth",
    # "django.contrib.contenttypes",
    # "django.contrib.sessions",
    # "django.contrib.messages",
    # "hera_django",
    # "hera_formsng",
    # "hera_skin",
    # "hera_utils",
    #END:installed_apps

    "djira",
    "djira.tests",
)

SITE_ID = 1

MIDDLEWARE_CLASSES = (
    # "django.contrib.sessions.middleware.SessionMiddleware",
    # "django.contrib.auth.middleware.AuthenticationMiddleware",
    # "django.contrib.messages.middleware.MessageMiddleware",
)

ROOT_URLCONF = "djira.test_urls"

TEMPLATES = ({
    "BACKEND": "django.template.backends.django.DjangoTemplates",
    "APP_DIRS": True,
    "OPTIONS": {
        "context_processors": (
            'django.contrib.messages.context_processors.messages',
        ),
    },
}, )



# -*- coding: utf-8 -*-

# $Id:$

from __future__ import print_function
from __future__ import unicode_literals
from future.utils import PY3

import unittest
if PY3:
    from unittest import mock
else:
    import mock

from django.utils.datastructures import MultiValueDict

from .. import validators as V
from ..schema import Bool
from ..schema import Float
from ..schema import Int
from ..schema import List
from ..schema import Schema
from ..schema import String
from ..schema import _Type
from ..schema import _UNDEFINED


class _BaseTestCase(unittest.TestCase):

    def assertConversion(self, value, expected, type_=None):
        if type_ is None:
            type_ = self.type_
        self.assertEqual(type_.to_python(value), expected)

    def assertValidationFails(self, value, regex=None, type_=None):
        if regex is None:
            manager = self.assertRaises(V.SchemaError)
        else:
            manager = self.assertRaisesRegexp(V.SchemaError, regex)
        if type_ is None:
            type_ = self.type_
        with manager:
            type_.to_python(value)


class Test_Type(_BaseTestCase):

    def mock_python_type(self, type_, **kwargs):
        return mock.patch.object(type_, "python_type", **kwargs)

    # to_python

    ## managing undefined values

    def test_to_python_UNDEFINED_with_no_default_raises_SchemaError(self):
        type_ = _Type()
        with self.assertRaisesRegexp(V.SchemaError, r"missing required value"):
            type_.to_python(_UNDEFINED)

    def test_to_python_UNDEFINED_performs_convertion_on_default(self):
        type_ = _Type(default=42)
        with self.mock_python_type(type_) as python_type:
            type_.to_python(_UNDEFINED)
            python_type.assert_called_once_with(42)

    def test_to_python_UNDEFINED_runs_validators_on_default(self):
        type_ = _Type(default=42, validators=[mock.Mock()])
        with mock.patch.object(type_, "_apply_validators") as apply_validators:
            with self.mock_python_type(type_, return_value=314):
                type_.to_python(_UNDEFINED)
                apply_validators.assert_called_once_with(314)

    ## initial type conversion

    def test_to_python_calls_self_dot_python_type(self):
        type_ = _Type()
        with self.mock_python_type(type_, return_value=314) as pt:
            self.assertEqual(type_.to_python(42), 314)
            pt.assert_called_once_with(42)

    def test_to_python_converts_TypeError_to_SchemaError(self):
        type_ = _Type()
        with self.mock_python_type(type_, side_effect=TypeError("foo")):
            with self.assertRaisesRegexp(V.SchemaError, r"foo"):
                type_.to_python(42)

    def test_to_python_converts_ValueError_to_SchemaError(self):
        type_ = _Type()
        with self.mock_python_type(type_, side_effect=ValueError("foo")):
            with self.assertRaisesRegexp(V.SchemaError, r"foo"):
                type_.to_python(42)

    def test_to_python_ignores_other_exceptions(self):
        type_ = _Type()
        with self.mock_python_type(type_, side_effect=ZeroDivisionError("foo")):
            with self.assertRaisesRegexp(ZeroDivisionError, r"foo"):
                type_.to_python(42)

    ## extra validators

    def test_to_python_chains_python_type_with__apply_validators(self):
        type_ = _Type()
        with mock.patch.object(type_, "_apply_validators", return_value=12345) as av:
            with self.mock_python_type(type_, return_value=314) as pt:
                self.assertEqual(type_.to_python(42), 12345)
                pt.assert_called_once_with(42)
                av.assert_called_once_with(314)

    # _apply_validators

    def test_apply_validators_chains_results(self):
        validators = [
            mock.Mock(return_value=23),
            mock.Mock(return_value=42),
        ]
        type_ = _Type(validators=validators)
        type_._apply_validators("asdf")
        validators[0].assert_called_once_with("asdf")
        validators[1].assert_called_once_with(23)

    def test_apply_validators_returns_last_result(self):
        validators = [
            mock.Mock(return_value=23),
            mock.Mock(return_value=42),
        ]
        type_ = _Type(validators=validators)
        self.assertEqual(type_._apply_validators("asdf"), 42)

    def test_validators_chain_stops_on_first_error(self):
        validators = [
            mock.Mock(return_value=23),
            mock.Mock(side_effect=V.SchemaError),
            mock.Mock(return_value=42),
        ]
        type_ = _Type(validators=validators)
        with self.assertRaises(V.SchemaError):
            type_._apply_validators("asdf")
        self.assertEqual(validators[0].call_count, 1)
        self.assertEqual(validators[1].call_count, 1)
        self.assertEqual(validators[2].call_count, 0)


class TestBool(_BaseTestCase):

    def setUp(self):
        self.type_ = Bool()

    def test_initial_conversion(self):
        self.assertConversion("true", True)
        self.assertConversion(True, True)
        self.assertConversion("false", False)
        self.assertConversion(False, False)
        self.assertValidationFails("True", "not boolean.*")
        self.assertValidationFails("False", "not boolean.*")
        self.assertValidationFails("0", "not boolean.*")
        self.assertValidationFails("1", "not boolean.*")
        self.assertValidationFails("a", "not boolean.*")


class TestInt(_BaseTestCase):

    def setUp(self):
        self.type_ = Int()

    def test_initial_conversion(self):
        self.assertConversion("0", 0)
        self.assertConversion("42", 42)
        self.assertConversion("-314", -314)
        self.assertValidationFails("a")
        self.assertValidationFails("3.14")


class TestFloat(_BaseTestCase):

    def setUp(self):
        self.type_ = Float()

    def test_initial_conversion(self):
        self.assertConversion("0", 0)
        self.assertConversion("42", 42)
        self.assertConversion("-314", -314)
        self.assertConversion("3.14", 3.14)
        self.assertValidationFails("a")


class TestList(_BaseTestCase):

    def test_initial_conversion(self):
        self.assertConversion(["1", "3"], [1, 3], type_=List(Int()))

    def test_undefined_list_without_default_raise_SchemaError(self):
        type_ = List(Int())
        with self.assertRaisesRegexp(V.SchemaError, r"missing required value"):
            type_.to_python(_UNDEFINED)

    def test_undefined_list_with_default_returns_default(self):
        type_ = List(Int(), default=[42, 314])
        self.assertEqual(type_.to_python(_UNDEFINED), [42, 314])

    def test_calls_item_type_dot_to_python_for_each_item(self):
        to_python = mock.Mock(wraps=lambda x: int(x) * 2)
        item_type = _Type()
        item_type.to_python = to_python
        type_ = List(item_type)

        # just to be sure that mocking hasn't broken anything
        self.assertConversion(["1", "-3", "42"], [2, -6, 84], type_=type_)

        self.assertEqual(to_python.call_count, 3)
        self.assertEqual(
            to_python.call_args_list,
            [(("1", ), ), (("-3", ), ), (("42", ), )]
        )

    def test_applies_validators_to_the_resulting_list(self):
        v = mock.Mock(return_value="whatever")
        type_ = List(Int(), validators=[v])
        self.assertEqual(type_.to_python(["1", "3"]), "whatever")
        v.assert_called_once_with([1, 3])


class TestSchema(_BaseTestCase):

    def setUp(self):
        self.type_ = Schema(
            schema=dict(
                name=String(),
                value=Int(),
                items=List(String(), default=[])
            )
        )
        self.type_with_default = Schema(
            schema=dict(
                name=String(),
                value=Int(),
                items=List(String(), default=["foo", "bar"])
            ),
            default=dict(name="Bob", value=42),
        )

    def mk_dict(self, **kwargs):
        res = MultiValueDict()
        for k, v in kwargs.items():
            if isinstance(v, list):
                res.setlist(k, v)
            else:
                res[k] = v
        return res

    # __init__

    def test_init_converts_default_to_MultiValueDict(self):
        self.assertIsInstance(self.type_with_default.default, MultiValueDict)

    def test_MultiValueDict_conversion(self):
        d = {"the answer": 42, "hello": "world", "a_list": ["of", "things"]}
        res = Schema._make_multidict(d)
        self.assertIsInstance(res, MultiValueDict)
        self.assertEqual(res.get("the answer"), 42)
        self.assertEqual(res.get("hello"), "world")
        self.assertEqual(res.get("a_list"), "things")
        self.assertEqual(res.getlist("a_list"), ["of", "things"])

    # to_python

    def test_undefined_dict_without_default_raises_SchemaError(self):
        with self.assertRaisesRegexp(V.SchemaError, r"missing required value"):
            self.type_.to_python(_UNDEFINED)

    def test_undefined_dict_with_default_returns_default_augmented(self):
        self.assertNotIn("default", self.type_with_default.default)
        self.assertEqual(
            self.type_with_default.to_python(_UNDEFINED),
            {"name": "Bob", "value": 42, "items": ["foo", "bar"]}
        )

    def test_raises_SchemaError_if_value_contains_unknown_fields(self):
        value = self.mk_dict(
            name="Bob", value="3", unexpected="not in the schema"
        )
        with self.assertRaisesRegexp(V.SchemaError, r"Unknown field.*"):
            self.type_.to_python(value)

    def test_initial_conversion(self):
        value = self.mk_dict(name="Bob", value="3")
        self.assertConversion(
            value,
            {"name": "Bob", "value": 3, "items": []},
        )

    def test_validates_individual_keys(self):
        value = self.mk_dict(name="Bob", value="x")
        self.assertValidationFails(value, r"value: invalid literal.*")

    def test_applies_validators_to_the_resulting_dict(self):
        value = self.mk_dict(name="Bob", value="3")
        v = mock.Mock(return_value="whatever")
        self.type_._validators = [v]
        self.assertEqual(self.type_.to_python(value), "whatever")
        v.assert_called_once_with({"name": "Bob", "value": 3, "items": []})



# -*- coding: utf-8 -*-

# $Id:$

from importlib import import_module

from django.apps import AppConfig
from django.core.exceptions import ImproperlyConfigured
import pluggy

from . import hookspec
from .common import get_config_value
from .common import PROJECT_NAME


class DjiraAppConfig(AppConfig):
    name = PROJECT_NAME
    verbose_name = PROJECT_NAME

    def ready(self):
        pm = pluggy.PluginManager(PROJECT_NAME)
        self.plugin_manager = pm
        pm.add_hookspecs(hookspec)
        loader = self._get_plugin_loader()
        loader(pm)

        pm.hook.initialize()

    def _get_plugin_loader(self):
        loader_fqn = get_config_value("plugin_loader", "djira.plugin_loader.entry_points")
        module_name, loader_name = loader_fqn.rsplit(".", 1)
        module = import_module(module_name)
        if not hasattr(module, loader_name):
            raise ImproperlyConfigured(
                "DJIRA.plugin_loader: loader not found '{}'".format(loader_name)
            )
        return getattr(module, loader_name)



# -*- coding: utf-8 -*-

# $Id: __init__.py 1251 2016-03-25 14:10:28Z alex $

import traceback

from django.conf import settings
from django.http.response import HttpResponseForbidden
from django.http.response import HttpResponseNotFound
from django.http.response import JsonResponse
from django.views.decorators.http import require_http_methods

from .common import logger
from .ep_registry import get_endpoint_by_name
from .validators import SchemaError


@require_http_methods(["GET", "HEAD"])
def dispatcher(request, name):
    """Dispatches the call to the endpoint.

    """
    if not settings.DEBUG:
        return HttpResponseForbidden()

    # get endpoint
    ep = get_endpoint_by_name(name)
    if ep is None:
        return HttpResponseNotFound(name)

    # validate arguments
    if ep.request_schema:
        try:
            params = ep.request_schema.to_python(request.GET)
        except SchemaError as e:
            logger.info("Error in arguments: %s", e.args[0])
            return _error_response(400, e.args[0])
        except Exception as e:
            logger.error(
                "Unexpected error validating arguments: endpoint='%s', GET=%r",
                name, request.GET,
                exc_info=True
            )
            return _error_response(500, e.args[0], details=e)
    else:
        params = {}

    # run endpoint
    try:
        res = ep.function(**params)
    except Exception as e:
        logger.error(
            "Unexpected error running endpoint: endpoint='%s', kwargs=%r",
            name, params,
            exc_info=True
        )
        return _error_response(500, e.args[0], details=e)

    return JsonResponse(res, encoder=ep.encoder, safe=False)


#             _            _
#  _ __  _ __(_)_   ____ _| |_ ___
# | '_ \| '__| \ \ / / _` | __/ _ \
# | |_) | |  | |\ V / (_| | ||  __/
# | .__/|_|  |_| \_/ \__,_|\__\___|
# |_|


def _error_response(http_status, description, error_code=None, details=None):
    if isinstance(details, Exception):
        details = traceback.format_exc()
    response = {
        "message": description,
        "code": error_code,
        "details": details,
    }
    return JsonResponse(response, status=http_status)



# -*- coding: utf-8 -*-

# $Id:$

"""Endpoint registry.

This module provides interaction with the endpoints registry:

- initialization: calls ``get_endpoints`` on each registered plugin
  and populates the registry. It takes care of handling name
  collisions.

- querying: allows to check the registered endpoints.

"""

from __future__ import print_function
from __future__ import unicode_literals

import itertools

from .common import EndPoint
from .common import get_plugin_manager
from .schema import get_schema_spec


# endpoints registry
_endpoints = None


def get_endpoint_by_name(name):
    """Returns an endpoint given a name.

    If theres no endpoint named ``name`` returns ``None``.

    """
    global _endpoints

    if _endpoints is None:
        pm = get_plugin_manager()
        _endpoints = {i.name: i for i in _DEFAULT_ENDPOINTS}
        _update_endpoints(_endpoints, pm.hook.get_endpoints())

    return _endpoints.get(name, None)

#             _            _
#  _ __  _ __(_)_   ____ _| |_ ___
# | '_ \| '__| \ \ / / _` | __/ _ \
# | |_) | |  | |\ V / (_| | ||  __/
# | .__/|_|  |_| \_/ \__,_|\__\___|
# |_|

# builtin endpoints


def ep_list():
    """List available endpoints.

    """

    res = []
    for ep in sorted(_endpoints.values(), key=lambda x: x.name):
        if ep.request_schema:
            args_doc = get_schema_spec(ep.request_schema)
        else:
            args_doc = {}
        res.append({
            "name": ep.name,
            "description": ep.doc,
            "parameters": args_doc,
        })
    return res


def ep_ping():
    """Check service availability.

    """
    return "pong"


def ep_version():
    """Get djira API version.

    This endpoint returns the ``djira`` API version so that clients
    can check their compatibility.

    .. warning:: Don't confuse this API with the one provided by the
                 actual endpoints. The version returned here refers to
                 the format of the messages, the specification of the
                 builtin endpoints etc. Ideally third party plugin
                 will provide it's own version.

    Returns a list with two integers representing the djira API
    version, using the format ``[major, minor]``:

    - ``minor`` will be incremented whenever a non breaking change is
      made. For example, adding a new builtin endpoint. Older clients
      should ignore those elements.

    - ``major`` will be incremented when a change breaks compatibility
      with existing clients.

    """
    # TODO: move the actual version "value" somewhere else?
    return [1, 0]


_DEFAULT_ENDPOINTS = [
    EndPoint(ep_list, name="__list__", doc="List available endpoints."),
    EndPoint(ep_ping, name="__ping__", doc="Test service availability."),
    EndPoint(ep_version, name="__version__", doc="Get djira API version."),
]


# helper functions

def _update_endpoints(res, l):
    for ep in itertools.chain(*l):
        if not isinstance(ep, EndPoint):
            raise ValueError("not an endpoint {!r}".format(ep))
        if ep.name in res:
            raise ValueError("endpoint already exists: {}".format(ep.name))
        res[ep.name] = ep



# -*- coding: utf-8 -*-

# $Id:$

from __future__ import print_function
from __future__ import unicode_literals

import inspect

from django.apps import apps

from ..common import EndPoint
from ..common import hookimpl
from .. import schema as S


@hookimpl
def get_endpoints():
    return (
        EndPoint(get_apps_list),
        EndPoint(get_apps_details, request_schema=get_apps_details_schema),
    )


def get_apps_list():
    """Returns a list with the labels of the installed apps.

    """
    return sorted(apps.app_configs.keys())


get_apps_details_schema = S.Schema(
    schema=dict(
        labels=S.List(S.String(doc="app label"))
    )
)


def get_apps_details(labels):
    """Returns details about the given apps.

    """
    res = {}
    for app_label in labels:
        try:
            app = apps.get_app_config(app_label)
        except LookupError:
            details = None
        else:
            details = _get_app_detail(app)
        res[app_label] = details
    return res


def _get_app_detail(app):
    return {
        "label": app.label,
        "models": _get_models(app),
        "name": app.name,
        "path": app.path,
        "app_class_name": app.__class__.__name__,
        "app_class_source": inspect.getsourcefile(app.__class__),
        "app_class_line": inspect.getsourcelines(app.__class__)[1],
        "verbose_name": app.verbose_name,
    }


def _get_models(app):
    return sorted([
        class_._meta.object_name
        for class_ in app.models.values()
    ])



# -*- coding: utf-8 -*-

# $Id:$

"""The schema allows to specify formally the call interface for an
endpoint.

Example/Motivation: TODO

How missing/undefined values work: TODO


"""

from __future__ import print_function
from __future__ import unicode_literals

import functools

from django.utils.datastructures import MultiValueDict

from . import validators as V

# marker for missing values
_UNDEFINED = object()


def handle_undefined_value(f):
    @functools.wraps(f)
    def decorator(self, value):
        if value is _UNDEFINED:
            if self.default is _UNDEFINED:
                raise V.SchemaError("missing required value")
            else:
                # NOTE: just to be safe, assume that the default value
                # may contain errors
                value = self.default

        return f(self, value)
    return decorator


class _Type(object):
    """Base class for schema types.

    """

    type_name = None
    python_type = None

    def __init__(self, default=_UNDEFINED, doc="", validators=()):
        self.default = default
        self.doc = doc
        self._validators = list(validators)  # make a copy

    @handle_undefined_value
    def to_python(self, value):
        """Converts a value from the request to a python type.

        """
        try:
            value = self.python_type(value)
        except (ValueError, TypeError) as e:
            raise V.SchemaError(*e.args)

        return self._apply_validators(value)

    def _apply_validators(self, value):
        for validator in self._validators:
            value = validator(value)
        return value


class Bool(_Type):
    """Boolean values.

    It's very strict regarding the allowed values, only the strings
    ``"true"`` and ``"false"``, in lowercase, and the python's boolean
    literals ``True`` and ``False`` are recognized, everything else
    raises an ``SchemaError`` exception.

    """
    type_name = "boolean"

    @staticmethod
    def python_type(v):
        if v in (True, "true"):
            return True
        if v in (False, "false"):
            return False
        raise TypeError("not boolean: {!r}".format(v))


class Int(_Type):
    type_name = "int"
    python_type = int


class Float(_Type):
    type_name = "float"
    python_type = float


class String(_Type):
    type_name = "string"
    python_type = str


class List(_Type):
    type_name = "list"
    python_type = list

    def __init__(self, item_type, *args, **kwargs):
        super(List, self).__init__(*args, **kwargs)
        self._item_type = item_type

    @handle_undefined_value
    def to_python(self, value):
        to_python = self._item_type.to_python
        res = [to_python(i) for i in value]
        return self._apply_validators(res)


class Schema(_Type):
    type_name = "schema"
    python_type = dict

    def __init__(self, schema, *args, **kwargs):
        if "default" in kwargs:
            kwargs["default"] = self._make_multidict(kwargs["default"])
        super(Schema, self).__init__(*args, **kwargs)
        self._schema = dict(schema)  # make a copy

    @handle_undefined_value
    def to_python(self, value):
        fields_received = set(value.keys())
        fields_expected = set(self._schema.keys())
        unknown_fields = fields_received.difference(fields_expected)
        if unknown_fields:
            raise V.SchemaError(
                "Unknown field(s): {!r}".format(sorted(unknown_fields))
            )

        res = {}
        for name, field in self._schema.items():
            if isinstance(field, List):
                v = value.getlist(name, _UNDEFINED)
            else:
                v = value.get(name, _UNDEFINED)

            try:
                v = field.to_python(v)
            except V.SchemaError as e:
                args = list(e.args)
                args[0] = "{}: {}".format(name, args[0])
                e.args = tuple(args)
                raise e
            else:
                res[name] = v

        return self._apply_validators(res)

    @staticmethod
    def _make_multidict(d):
        res = MultiValueDict()
        for k, v in d.items():
            if isinstance(v, list):
                res.setlist(k, v)
            else:
                res[k] = v
        return res


def get_schema_spec(item):
    """Returns a dictionary with the Schema's specification.

    """
    return _get_schema_spec(item, None)


def _get_schema_spec(item, parent=None):
    doc = {
        "description": item.doc,
        "type": item.type_name,
    }
    if item.default is _UNDEFINED:
        doc["required"] = True
    else:
        doc["required"] = False
        doc["default"] = item.default

    if isinstance(item, List):
        doc["extra"] = _get_schema_spec(item._item_type, item)
    elif isinstance(item, Schema):
        doc["extra"] = {k: _get_schema_spec(v, item) for k, v in item._schema.items()}

    if not isinstance(parent, Schema):
        # NOTE: _Type defines "required" and "default" for convenience
        # but they only make sense for items within an Schema (any
        # named container).
        doc.pop("required", None)
        doc.pop("default", None)

    return doc



# -*- coding: utf-8 -*-

# $Id:$

"""Plugin loaders.

A plugin loader implements the policy used to discover and load
plugins.

A loader it's just a callable that receives a ``plugin_manager`` as
argument and populates it with plugins.

"""

from __future__ import print_function
from __future__ import unicode_literals

from importlib import import_module

from .common import get_config_value
from .common import PROJECT_NAME


def entry_points(plugin_manager):
    """Load plugins from setuptools' entry points.

    Loads the plugins registered as ``djira`` entry points.

    """
    plugin_manager.load_setuptools_entrypoints(PROJECT_NAME)


def django_config(plugin_manager):
    """Load plugins listed in the configuration.

    Loads the plugins listed in the djira's configuration
    ``enabled_plugins``. Each entry in the list is the name (an
    string) of a module.

    """
    for module_name in get_config_value("enabled_plugins", []):
        module = import_module(module_name)
        plugin_manager.register(module)



# -*- coding: utf-8 -*-

# $Id:$

from __future__ import print_function
from __future__ import unicode_literals

import logging

from django.apps import apps
from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder

import pluggy


logger = logging.getLogger("djira")


# Project name, required by pluggy. Used as the name of the app too.
PROJECT_NAME = "djira"


class EndPoint(object):
    """Class describing an endpoint.

    :param callable function: the function that will be called for the
      endpoint. Will receive the arguments from the GET request
      payload. No ``request`` will be passed, it's not a view. Return
      an object that can be serialized to json.

    :param str name: name of the endpoint, used for the URL. If
      omitted ``function.__name__`` will be used.

    :param Schema request_schema: ``Schema`` describing the parameters
      accepted by ``function``. Leave blank if ``function`` takes no
      arguments. Is used both for sanitizing the input and for
      documentation purposes.

    :param str doc: documentation for of the endpoint. If omitted
      ``function.__doc__`` will be used.

    :param object encode: json encode used to convert the value
      returned by ``function`` to json. Specify a value in case
      ``function`` returns some non *standard* value.

    """
    def __init__(self, function, name=None, request_schema=None,
                 doc=None, encoder=DjangoJSONEncoder):
        if not callable(function):
            raise TypeError("'function' argument must be callable.")
        if name is None:
            name = function.__name__
        if doc is None:
            doc = function.__doc__
        self.function = function
        self.name = name
        self.request_schema = request_schema
        self.doc = doc
        self.encoder = encoder


def get_plugin_manager():
    """Return the plugin manager.

    """
    app = apps.get_app_config(PROJECT_NAME)
    return app.plugin_manager


def get_config_value(name, default=None):
    """Return the value for a djira configuration option.

    """
    DJIRA = getattr(settings, "DJIRA", {})
    return DJIRA.get(name, default)


# decorator required by pluggy in order to mark plugin implementation
# functions.
hookimpl = pluggy.HookimplMarker(PROJECT_NAME)



# -*- coding: utf-8 -*-

# $Id:$

import codecs
import os

from setuptools import setup, find_packages


version = "0.2.1"


def read(*path_components):
    with codecs.open(os.path.join(*path_components), "r") as f:
        return f.read()


setup(
    name="djira",
    version=version,
    description="Django Introspection REST API.",
    long_description=read("README.rst"),
    classifiers=[
        "Development Status :: 1 - Planning",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 1.9",
        "Framework :: Django :: 1.10",
        "Framework :: Django :: 1.11",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development",
    ],
    keywords="",
    author="Alexis Roda",
    author_email="alexis.roda.villalonga@gmail.com",
    url="https://github.com/patxoca/djira",
    license="GPL",
    packages=find_packages(exclude=["ez_setup"]),
    namespace_packages=[],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "setuptools",
        "django",
        "pluggy",
    ],
    entry_points={
    },
)



# -*- coding: utf-8 -*-

# $Id:$

from __future__ import print_function
from __future__ import unicode_literals

import functools
import inspect

from django.urls.resolvers import RegexURLPattern
from django.urls.resolvers import RegexURLResolver
from django.urls import get_resolver
from django.urls import get_urlconf

from ..common import EndPoint
from ..common import hookimpl


@hookimpl
def get_endpoints():
    return (
        EndPoint(get_urls_details),
    )


def get_urls_details():
    """Return a list with the models names."""
    patterns = []
    _get_url_patterns(get_resolver(get_urlconf()), patterns)
    res = {}
    for i in patterns:
        info = {
            "url_name": i.name,
        }
        info.update(_get_callback_info(i.callback))
        res[i.name] = info
    return res


def _get_url_patterns(resolver, result):
    for i in resolver.url_patterns:
        if isinstance(i, RegexURLPattern):
            result.append(i)
        elif isinstance(i, RegexURLResolver):
            _get_url_patterns(i, result)
        else:
            raise TypeError(type(i))


def _get_callback_info(cb):
    cb_name = None
    cb_path = None
    cb_lineno = None
    if hasattr(cb, "__name__"):
        cb_name = cb.__name__
    while True:
        if hasattr(cb, "__wrapped__"):
            cb = cb.__wrapped__
        elif hasattr(cb, "view_class"):
            cb = cb.view_class
        elif isinstance(cb, functools.partial):
            cb = cb.func
        else:
            break
    try:
        cb_path = inspect.getsourcefile(cb)
        cb_lineno = inspect.getsourcelines(cb)[1]
    except TypeError:
        pass
    return {
        "callback_name": cb_name,
        "callback_path": cb_path,
        "callback_lineno": cb_lineno,
    }
